"""
*****************************************************************

Licensed Materials - Property of IBM

(C) Copyright IBM Corp. 2018, 2019. All Rights Reserved.

US Government Users Restricted Rights - Use, duplication or
disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

*****************************************************************
"""

import os
import ddlrun.error as error
import ddlrun.utilities as utilities


def local_conda_test():
    """Verifies that conda is installed and accessible locally.

    Returns:
        conda setup script path

    Raises:
        GenericDDLError
    """
    conda_exe_path = os.environ.get("CONDA_EXE")
    conda_setup_script = None

    if conda_exe_path is None:
        raise error.GenericDDLError(error.ErrorMessage.NO_LOCAL_CONDA_INSTALL)

    conda_setup_script = os.path.dirname(os.path.dirname(conda_exe_path))
    conda_setup_script += "/etc/profile.d/conda.sh"

    if conda_setup_script is None or not os.path.isfile(conda_setup_script):
        raise error.GenericDDLError(
            error.ErrorMessage.NO_LOCAL_CONDA_SETUP_SCRIPT)

    return conda_setup_script


def verify_host_software(args):
    """Verify that software is correctly configured on each host.
        * Checks that 'ddl' is installed on each node. For now do this by checking that
          'ddlrun' executes on each node.
        * Confirms that the locally active conda environment exists on each host.
        * Checks that the specified executable exists and is executable on each host.

    Args:
        args: Parsed program arguments.

    Returns:
        None

    Raises:
        DDLError
    """
    encountered_error = False
    for host in args.hosts:
        if not args.ignore_conda_env:
            # Check conda install is same on all hosts.
            # If the source script doesn't exist in the same place on all hosts,
            # then conda is not installed in the same place on all the hosts (which
            # is a problem)
            ret_code, remote_conda_setup_script, stderr = utilities.launcher_run_command(
                "ls %s" % (args.conda_setup_script), host, args=args)
            if not ret_code:
                encountered_error = True
                error.log_info(stderr)
                error.log_info(remote_conda_setup_script)
                error.log_error(
                    error.ErrorMessage.NO_REMOTE_CONDA_SETUP_SCRIPT, host,
                    args.conda_setup_script)
                continue

            # Check local activated conda environment exists on all remote hosts
            ret_code, remote_conda_env, stderr = utilities.launcher_run_command(
                "ls %s" % (args.conda_env), host, args=args)
            if not ret_code:
                encountered_error = True
                error.log_info(stderr)
                error.log_info(remote_conda_env)
                error.log_error(error.ErrorMessage.NO_REMOTE_CONDA_ENV,
                                args.conda_env, host)
                continue

        # Check for remote ddlrun.
        ret_code, result, stderr = utilities.launcher_run_command(
            "ddlrun --help", host, conda_env=args.conda_env, args=args)
        error.log_info(stderr)
        if not ret_code:
            encountered_error = True
            error.log_error(error.ErrorMessage.SW_ERROR, host)
            continue

        # Check for remote executable.
        ret_code, result, stderr = utilities.launcher_run_command(
            "bash -c \"[ -x \\\"$(which %s)\\\" ]\"" % (args.program[0]),
            host,
            conda_env=args.conda_env,
            args=args)
        error.log_info(result)
        error.log_info(stderr)
        if not ret_code:
            encountered_error = True
            error.log_error(error.ErrorMessage.NO_EXECUTABLE, args.program[0],
                            host)
            continue
    if encountered_error:
        raise error.DDLError()


def get_conda_env():
    """Gets the active conda environment of the current system.

    Returns:
        Name of the active conda environment.

    Raises:
        GenericDDLError
    """
    conda_env_path = os.environ.get("CONDA_PREFIX")
    if conda_env_path is None:
        raise error.GenericDDLError(error.ErrorMessage.NO_LOCAL_CONDA_ENV)

    return conda_env_path
