"""
*****************************************************************

Licensed Materials - Property of IBM

(C) Copyright IBM Corp. 2018, 2019. All Rights Reserved.

US Government Users Restricted Rights - Use, duplication or
disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

*****************************************************************
"""
import logging
import subprocess
import os
import re
import socket
import psutil
import yaml

import ddlrun.error as error

class Launcher:  #pylint: disable=no-init,too-few-public-methods
    """Enum for Launchers."""
    JSRUN = 0
    MPIRUN = 1

# The order of LAUNCHERS determines the priority of default launcher.
# The first launcher in the list is the highest priority default.
LAUNCHERS = {
    Launcher.MPIRUN:
    "mpirun",
    Launcher.JSRUN: #JSRUN should be made default once the fix for the arg parse bug is available.
    "jsrun"
}

# These ssh options force the usage of private/public keys and don't prompt the user for input.
SSH_ARGS = {
    "-oPasswordAuthentication=no", "-oStrictHostKeyChecking=no",
    "-oBatchMode=yes"
}

def run_command_capture(cmd):
    """Run a shell command and capture the ret_code, stdout and stderr."""
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
        universal_newlines=True)
    output, error_out = process.communicate()

    return process.returncode == 0, output, error_out


def run_command(cmd):
    """Run a shell command and don't collect the output."""
    process = subprocess.Popen(cmd, shell=True, universal_newlines=True)
    process.wait()
    return process.returncode

def run_program_with_ssh(cmd, args):
    """Run each instance of the program with a SSH call."""
    return_code = 0
    processes = []
    num_accelerators = int(args.topology.split("x")[0])
    group_id = get_group_id()
    index = 0
    working_dir = os.getcwd()
    for host in args.hosts:
        local_host = NetUtils.is_local_host(host)
        for gpu in range(num_accelerators):
            ddl_options = get_flex_ddl_options(args.num_procs, index, gpu, group_id, args.ddl_args)
            cmd_to_run = "%s %s bash -c '%s'" % \
                         (ddl_options, make_arg_list(args.env_vars), cmd)
            if not local_host:
                cmd_to_run = "ssh %s %s \"cd %s > /dev/null 2>&1; %s\"" % \
                             (make_arg_list(SSH_ARGS), host, working_dir, cmd_to_run)
            logging.getLogger("ddlrun").critical("+ %s", cmd_to_run)
            process = subprocess.Popen(cmd_to_run, shell=True, universal_newlines=True)
            processes.append(process)
            index += 1
    for process in processes:
        process.wait()
        # Return only the first non-zero return code encountered
        if return_code == 0:
            return_code = process.returncode
    return return_code

def append_all_text(filename, text):
    """Appends to a text file, creating it if it doesn't exist."""
    with open(filename, "a+") as output_file:
        output_file.write(text)


def read_all_text(filename):
    """Reads all test from a text file."""
    with open(filename, 'r') as text_file:
        text = text_file.read()
    return text


def make_arg_list(arglist):
    """Converts a list of strings to a an arg list."""
    return ' '.join(str(arg) for arg in arglist)

def quote_arg(arg):
    """Properly quotes a command line arg."""
    quotable_chars = set(" !\"#$&()*,;<>?^`{|}")
    if any((c in quotable_chars) for c in arg):
        return '"' + str(arg) + '"'
    return str(arg)

def make_prog_arg_list(arglist):
    """Converts a list of strings to a properly quoted arg list."""
    return ' '.join(quote_arg(arg) for arg in arglist)


def make_env_var_exports(env_vars):
    """Creates a sequence of export commands for the list of env_vars."""
    return make_arg_list(("export %s ;" % env_var for env_var in env_vars))


def get_flex_ddl_options(count, index, dev_index, group_id, ddl_args):
    """Returns DDL options to use in flex mode."""
    return "DDL_OPTIONS='-ddl_worker_count %s -ddl_worker_index %s " \
           "-ddl_device_index %s -ddl_group_id %s %s'" % \
           (count, index, dev_index, group_id, make_arg_list(ddl_args))


def get_group_id():
    """Returns the group identifier"""
    import pyddl
    return pyddl.generate_group_id()


def make_temp_rankfile(host, name, args):
    """Makes a simple temporary rankfile.

    Args:
        host: The single host to put in the rankfile.
        name: The suffix to use for the rankfile.
        args: Parsed program arguments.

    Returns:
        The full rankfile path.
    """
    temp_rank_filename = os.path.join(args.working_dir, "temp_rankfile_%s" % (name))
    temp_rank_cmd = "rank_gen.py --aisles 1 --racks 1 --accelerators 1"
    if args.launcher == LAUNCHERS[Launcher.JSRUN]:
        temp_rank_cmd += " --jsrun"
    temp_rank_cmd += " " + host
    _, rankfile, _ = run_command_capture(temp_rank_cmd)
    with open(temp_rank_filename, "w") as temp_rank_file:
        temp_rank_file.write(rankfile)

    return temp_rank_filename


def launcher_run_command(command,
                         host=None,
                         rank_file=None,
                         conda_env=None,
                         args=None):
    """Runs the given command using the launcher.

    If a host is given, then a temporary rankfile is made to run on that host.
    Otherwise the given rankfile is used, if provided.

    Args:
        command: The command to launch via the launcher.
        host: The host to generate a rankfile for.
        rank_file: Explicit rank_file to use.
        conda_env: If passed in, activates the conda env before executing command.
        args: Parsed program arguments.

    Returns:
        The results of running the command with the launcher.
    """
    if rank_file is None and args is not None and host is not None:
        rank_file = make_temp_rankfile(host, host, args)
    command_to_run = make_launcher_command(command, args, rank_file, conda_env)
    return run_command_capture(command_to_run)


def make_launcher_command(command,
                          args,
                          rank_file=None,
                          conda_env=None):
    """Constructs the string launcher command.

    Args:
        command: The command to launch via distributed launcher.
        args: The arguments passed to ddlrun
        rank_file: The rankfile to use.
        conda_env: If passed in, activates the conda env before executing command.

    Returns:
        The full launcher command.
    """
    result = args.launcher_exe

    if args.launcher == LAUNCHERS[Launcher.MPIRUN]:
        if args.debug_timeout is not None:
            result += " -timeout " + str(args.debug_timeout)

        if args.mpi_args is not None:
            result += " " + make_arg_list(args.mpi_args)

        result += " -n 1"

    elif args.launcher == LAUNCHERS[Launcher.JSRUN]:
        if args.jsrun_args is not None:
            result += " " + make_arg_list(args.jsrun_args)
        if args.env_vars is not None:
            result = ("%s %s")%(make_env_var_exports(args.env_vars), result)

    if rank_file is not None:
        if args.launcher == LAUNCHERS[Launcher.JSRUN]:
            result += " --erf_input " + str(rank_file)
        elif args.launcher == LAUNCHERS[Launcher.MPIRUN]:
            result += " --rankfile " + str(rank_file)
    elif args.launcher == LAUNCHERS[Launcher.JSRUN]:
        result += " -n 1"

    result += " bash -c '"

    if conda_env is not None:
        result += "source %s && conda activate %s  > /dev/null 2>&1 && " % \
                (args.conda_setup_script, conda_env)

    result += " " + str(command) + "'"

    return result


def executable_exists(executable):
    """Returns True if executable exists and is an executable file."""
    ret_code, _, _ = run_command_capture("which %s" % (executable))
    return ret_code


def add_env_var(args, variable, value=None):
    """Adds environment variable assignments for both mpi and ssh."""
    if value is None:
        args.mpi_args.extend(["-x", variable])
        if variable in os.environ:
            args.env_vars.append("%s='%s'" % (variable, os.environ[variable].strip()))
    else:
        args.mpi_args.extend(["-x", quote_arg("%s=%s" % (variable, value))])
        args.env_vars.append("%s='%s'" % (variable, value))


class LSFUtils: # pylint: disable=too-few-public-methods
    """LSF Utilities"""
    _CSM_ALLOCATION_QUERY = "/opt/ibm/csm/bin/csm_allocation_query"
    _csm_allocation_info = dict()

    @staticmethod
    def using_lsf():
        """Returns True if LSF was used to start the current process."""
        return "LSB_MCPU_HOSTS" in os.environ

    @staticmethod
    def get_allocation_info():
        """Returns and sets the static CSM allocation info, compute nodes
           and isolated core count."""
        if not LSFUtils._csm_allocation_info:
            if not executable_exists(LSFUtils._CSM_ALLOCATION_QUERY):
                if not "LSB_MCPU_HOSTS" in os.environ:
                    raise error.GenericDDLError(error.ErrorMessage.NO_LSF_ENV_VAR)
                LSFUtils._csm_allocation_info["compute_nodes"] = list()
                LSFUtils.get_allocation_info()["isolated_cores"] = 0
                lsf_host_list = os.environ["LSB_MCPU_HOSTS"].strip()
                for host in re.finditer(r"(\S+)\s+(\d+)", lsf_host_list):
                    LSFUtils._csm_allocation_info["compute_nodes"].append(host.group(1))
            else:
                if not "LSB_JOBID" in os.environ:
                    raise error.GenericDDLError(error.ErrorMessage.NO_LSF_JOB_ENV_VAR)
                lsf_job_id = os.environ["LSB_JOBID"].strip()
                _, allocation_yaml, _ = \
                    run_command_capture("%s -j %s" % (LSFUtils._CSM_ALLOCATION_QUERY, lsf_job_id))
                LSFUtils._csm_allocation_info = yaml.safe_load(allocation_yaml)
            # "LSB_MCPU_HOSTS" is in an arbitrary ordering, and LSF hostnames are alphabetical.
            # Sorting them allows rankgen to generate the best rankfile.
            LSFUtils._csm_allocation_info["compute_nodes"].sort()
        return LSFUtils._csm_allocation_info

    @staticmethod
    def get_compute_hosts():
        """Returns the list of LSF compute hosts."""
        return LSFUtils.get_allocation_info()["compute_nodes"]

    @staticmethod
    def get_isolated_cores():
        """Returns the amount of cores isolated by CSM."""
        return LSFUtils.get_allocation_info()["isolated_cores"]


class NetUtils: # pylint: disable=too-few-public-methods
    """Network utilities"""
    _local_ips_cache = set()

    @staticmethod
    def is_local_host(hostname):
        """Returns True if hostname resolves to a local ip address"""
        if not NetUtils._local_ips_cache:
            NetUtils._local_ips_cache = {ip.address for _, info in psutil.net_if_addrs().items()
                                         for ip in info if ip.family == socket.AF_INET or
                                         ip.family == socket.AF_INET6}
        hostname_info = None
        try:
            hostname_info = socket.getaddrinfo(hostname, None, socket.AF_INET)
        except socket.gaierror:
            # Check if there is an IPV6 address associated to this hostname instead
            try:
                hostname_info = socket.getaddrinfo(hostname, None, socket.AF_INET6)
            except socket.gaierror:
                # Keep going if hostname could not be found, because we will test connectivity later
                pass
        if hostname_info:
            # structure
            #[(family, socktype, proto, canonname, sockaddr)...] with sockaddr being (address, ...)
            return hostname_info[0][4][0] in NetUtils._local_ips_cache
        return False
